import json
import requests
from web3 import Web3
from eth_account import Account
from web3.gas_strategies.time_based import construct_time_based_gas_price_strategy
from web3.middleware import geth_poa_middleware
from xy_wallet.utils.BaseAPI import BaseConfig
import os

directory = os.path.dirname(__file__)


class BaseWalletAPI:
    def __init__(self,
                 init_account=False,
                 network_url=None,
                 poa=False,
                 keystore_dir=None,
                 keystore_filename=None,
                 pwd=123
                 ):
        self.pwd = pwd
        self.ConfigUtil = BaseConfig.ConfigUtil()

        # 装载配置文件
        self.get_config(
            keystore_filename,
            network_url,
            keystore_dir
        )

        # 查看有没有key_file
        try:
            with open(self.keystore_dir+self.keystore_filename, 'r') as r:
                data = r.read()
                if data is None or data == "":
                    init = True
                else:
                    init = False
        except Exception:
            init = True

        # 查看是否强制初始化账户
        init = init_account

        # 是否要初始化账户
        if init:
            self.init_account()
        else:
            self.load_account()

        # 是否启动POA链
        self.poa = poa
        if self.poa:
            self.web3.middleware_onion.inject(geth_poa_middleware, layer=0)

        self.web3 = Web3(Web3.HTTPProvider(self.network_url))

    @property
    def show(self):
        return {"address": self._address, "private": self._privatekey}

    @property
    def private_key(self):
        return self._privatekey

    @property
    def address(self):
        return self._address

    def get_config(self,
                   keystore_filename,
                   network_url,
                   keystore_dir
                   ):
        self.keystore_filename = self.ConfigUtil.get[
            'KEYNAME'] if keystore_filename is None else keystore_filename
        self.network_url = self.ConfigUtil.get['NETWORK'][0] if network_url is None else network_url

        def parse_path(path, num):
            for _ in range(1, num+1):
                path = path[:path.rfind("/")]
            return path

        self.keystore_dir = parse_path(directory, 2)+"/conf/keystores/"
        self.gas_url = self.ConfigUtil.get['GAS_URL']

    def update_network(self, url):
        self.network_url = url
        self.ConfigUtil.set('NETWORK', url)
        return True

    def to_checksum_address(self, addr):
        return self.web3.toChecksumAddress(addr)

    def create_wallet(self, pwd):
        with open(self.keystore_dir+self.keystore_filename, "w") as w:
            acc = Account.create()
            address = acc.address
            privatekey = acc.privateKey
            w.write(json.dumps(Account.encrypt(privatekey, pwd)))
            return {"address": address, "private": privatekey.hex()}

    def recover_wallet(self, private: str, pwd):
        if isinstance(private, int):  # 输入过来的私钥是int格式
            private = hex(private)
        with open(self.keystore_dir+self.keystore_filename, "w") as w:
            acc = Account.privateKeyToAccount(private)
            address = acc.address
            w.write(json.dumps(Account.encrypt(private, pwd)))
            return {"address": address, "private": private}

    def recover_m(self, mnemonic, pwd):
        with open(self.keystore_dir+self.keystore_filename, "w") as w:
            Account.enable_unaudited_hdwallet_features()
            acc = Account.create_with_mnemonic(mnemonic)[0]
            address = acc._address
            privatekey = acc._private_key.hex()
            w.write(json.dumps(Account.encrypt(privatekey, pwd)))
            return {"address": address, "private": privatekey}

    def get_wallet_from_file(self, keystore_file):
        with open(keystore_file, "r") as r:
            private = Account.decrypt(json.loads(r.read()), self.pwd).hex()
            address = Account.privateKeyToAccount(private).address
            return {address: private}

    def load_account(self):
        """
        通过file装载账户
        """
        with open(self.keystore_dir+self.keystore_filename, "r") as r:
            self._privatekey = Account.decrypt(
                json.loads(r.read()), self.pwd).hex()
        self._address = Account.privateKeyToAccount(
            self._privatekey).address

    def init_account(self):
        """
        初始化账户

        :param pwd: _description_
        """
        self.init_keystore_file()
        self.load_account()

    def init_keystore_file(self):
        """
        keystore初始化
        """
        with open(self.keystore_dir+self.keystore_filename, "w") as w:
            w.write(json.dumps(Account.encrypt(
                Account.create().privateKey, self.pwd)))

    def sign_and_send_raw(self, params, private_key):
        """
        签名发送函数

        :param params: _description_
        :param private_key: _description_
        :return: _description_
        """
        # 使用私钥签名
        signed_tx = self.web3.eth.account.signTransaction(
            params, private_key=private_key)
        # 发送交易，并获得交易hash
        tx_hash = self.web3.eth.sendRawTransaction(signed_tx.rawTransaction)
        # 返回交易hash
        print("tx_hash: {}".format(tx_hash.hex))
        return tx_hash.hex()

    def get_balance(self, address):
        """
        获取余额

        :param address: _description_
        :return: _description_
        """
        return float(self.web3.fromWei(
            self.web3.eth.get_balance(address),
            "Ether"))

    def transfer_eth(self, target_address, amount, private_key):
        """
        以太坊转账

        :param target_address: _description_
        :param amount: _description_
        :param private_key: _description_
        :return: _description_
        """
        address = Account.privateKeyToAccount(private_key).address

        # 获取用户的nonce
        nonce = self.get_nonce(address)
        # 构建交易数据包
        params = {
            'nonce': nonce,
            'to': target_address,
            'value': self.web3.toWei(amount, 'ether'),
            'gas': 210000,
            'gasPrice': self.web3.toWei('150', 'gwei'),
            # 'gasPrice': self.web3.toHex(self.web3.toWei(90, "Gwei")),
            'from': address,
        }
        return self.sign_and_send_raw(params, private_key)

    def contract_fun(self,
                     contractaddress,
                     abi_file: str,
                     private_key,
                     function_name,
                     args=None,
                     call=False):
        """
        合约函数

        :param contractaddress: _description_
        :param abi_file: _description_
        :param private_key: _description_
        :param function_name: _description_
        :param args: _description_, defaults to None
        :param call: _description_, defaults to False
        :return: _description_
        """
        with open(abi_file, "r", encoding="utf-8") as r:
            contract_abi = json.loads(r.read())

        address = self.web3.eth.account.from_key(private_key).address
        contractaddress = self.web3.toChecksumAddress(contractaddress)

        # 通过地址找到nonce
        nonce = self.get_nonce(address)

        # 定义合约对象
        contract = self.web3.eth.contract(
            address=contractaddress, abi=contract_abi)
        params = {'from': address, 'nonce': nonce}

        if not call:
            txn = eval("contract.functions.{}{}".format(function_name, args))
            self.web3.eth.set_gas_price_strategy(
                construct_time_based_gas_price_strategy(max_wait_seconds=60,
                                                        sample_size=10))
            gasprice = self.web3.eth.generateGasPrice()
            estimate_gas = txn.estimateGas(params)
            params.update({
                'gas': int(estimate_gas*1.1),
                'gasPrice': int(gasprice*1.5)
            })
            txn = txn.buildTransaction(params)
        elif args is None:
            return eval("contract.functions.{}().call()".format(function_name))
        else:
            return eval("contract.functions.{}(\"{}\").call()".format(function_name,
                                                                      args))

        receipt = self.sign_and_send_raw(txn, private_key)
        return self.web3.eth.wait_for_transaction_receipt(timeout=360, transaction_hash=receipt)

    def get_gas_price(self):
        headers = {
            "user-agent": "Mozilla/5.0 (Linux; Android 6.0; Nexus 5 Build/MRA58N) \
                AppleWebKit/537.36 (KHTML, like Gecko) Chrome/100.0.4896.127 Mobile Safari/537.36"
        }
        resp_data = json.loads(requests.get(
            self.gas_url, headers=headers).text)
        return int(resp_data['result'], 16)

    def get_nonce(self, address):
        nonce = self.web3.eth.getTransactionCount(address)
        return nonce




import os
from xy_wallet.utils.BaseAPI.BaseWallet import BaseWalletAPI
from xy_wallet.utils.BaseAPI.BaseConfig import ConfigUtil
from xy_wallet.utils.BaseAPI.BaseEvent import BaseEvent

ConfigUtil = ConfigUtil()
BaseEvent = BaseEvent()


def parse_path(path, num):
    for _ in range(1, num+1):
        path = path[:path.rfind("/")]
    return path


directory = os.path.dirname(__file__)
directory = parse_path(directory, 1)


class BaseTokenClass:
    def __init__(self):
        pass

    def help(self):
        """
        获取帮助
        """
        return """
        代币类
        sample: wallet token add weth 0xabc213
                wallet token balance weth
                wallet token send weth 0xabc123 100
        event只单独支持transfer监听
        """

    def add(self, token_name, addr):
        """
        增加代币

        :param token_name: _description_
        :param addr: _description_
        """
        return ConfigUtil.set(
            "TOKEN-LIST", NAME=token_name, ADDRESS=hex(addr))

    def balance(self, pwd, token_name="eth", addr=None):
        """
        获取余额

        :param addr: _description_, defaults to None
        :param token_name: _description_, defaults to "eth"
        """
        balance_api = BaseWalletAPI(pwd=pwd)
        addr = balance_api.address if addr is None else addr
        if token_name == "eth":
            print("您的ETH余额是: {}".format(balance_api.get_balance(addr)))
        else:
            token_list = ConfigUtil.get['TOKEN-LIST']
            for i in token_list:
                if i['NAME'] == token_name:
                    _addr = i['ADDRESS']
            decimals = 10**self._get_decimals(token_name, balance_api)
            print("您的{}余额是: {}".format(token_name, balance_api.contract_fun(
                _addr,
                directory+"/conf/abi/abi.json",
                balance_api.private_key,
                "balanceOf",
                balance_api.address,
                True)/decimals))

    def get_list(self):
        """
        代币列表

        :return: _description_
        """
        return ConfigUtil.get['TOKEN-LIST']

    def send(self, pwd, token_name, to, value):
        """
        发送交易

        :param token_name: _description_
        :param to: _description_
        :param value: _description_
        :return: _description_
        """
        send_api = BaseWalletAPI(pwd=pwd)
        to = send_api.to_checksum_address(to)

        if token_name == "eth":
            send_api.transfer_eth(to, value, send_api.private_key)
            return True

        if token_name == 'eth':
            return send_api.transfer_eth(to, value, send_api.private_key)
        else:
            token_list = ConfigUtil.get['TOKEN-LIST']
            for i in token_list:
                if i['NAME'] == token_name:
                    __address = i['ADDRESS']
            decimal = self._get_decimals(token_name, send_api)
            send_api.contract_fun(
                __address,
                directory+'/conf/abi/abi.json',
                send_api.private_key,
                'transfer',
                (to, int(float(value)*10**int(decimal))),
                False
            )
            # TODO 判断成功否
            return True

    def event(self):
        """
        获取event
        """
        BaseEvent.main_function()
        BaseEvent.get_transfer_events(10)

    def _get_decimals(self, token_name, get_decimals):
        """
        获取小数点

        :param token_name: _description_
        :return: _description_
        """
        if token_name == "eth":
            return 18
        else:
            token_list = ConfigUtil.get['TOKEN-LIST']
            for i in token_list:
                if i['NAME'] == token_name:
                    __address = i['ADDRESS']
            return get_decimals.contract_fun(
                __address,
                directory+'/conf/abi/abi.json',
                get_decimals.private_key,
                "decimals",
                None,
                True
            )



from xy_wallet.utils.BaseAPI.BaseConfig import ConfigUtil

ConfigUtil = ConfigUtil()


class BaseToolClass:
    def __init__(self):
        pass

    def current(self):
        """
        获取网络
        """
        return ConfigUtil.get['NETWORK']

    def set(self, url):
        """
        更换网络

        :param url: _description_
        """
        return ConfigUtil.set("NETWORK", url)



from xy_wallet.utils.BaseAPI.BaseConfig import ConfigUtil
from xy_wallet.utils import (
    WalletClass,
    ToolClass,
    TokenClass,
)
import fire


WalletClass = WalletClass.BaseWalletClass()
ToolClass = ToolClass.BaseToolClass()
TokenClass = TokenClass.BaseTokenClass()
config = ConfigUtil()


class Wallet:
    def __init__(self):

        self.help = """

        Usage

            wallet token/tool/wallet show/crate/recover ...

        Parameters

            token
                token类 add(添加代币)、balance(获取余额)、get_list(获取列表)、send(发送交易)、event(监听转账)
            tool
                tool类 current(获取当前网络)、set(更换网络)
            wallet
                wallet类 show(查看钱包)、crate(创建钱包)、recover(恢复钱包)

        """

    def token(self):
        return TokenClass

    def tool(self):
        return ToolClass

    def wallet(self):
        return WalletClass

    def help(self):
        return self.help


def main():
    fire.Fire(Wallet)


if __name__ == "__main__":
    main()



from xy_wallet.utils.BaseAPI.BaseWallet import BaseWalletAPI


class BaseWalletClass:
    def __init__(self):
        pass

    def help(self):
        """
        获取帮助
        """
        return """
        钱包类
        sample: wallet show
                wallet token balance weth
                wallet token send weth 0xabc123 100
        """

    def show(self, pwd):
        """
        查看钱包

        :param pwd: _description_
        :return: _description_
        """
        show_api = BaseWalletAPI(pwd=pwd)
        return show_api.show

    def crate(self, pwd):
        """
        创建钱包

        :param pwd: _description_
        :return: _description_
        """
        crate_api = BaseWalletAPI(init_account=True)
        return crate_api.create_wallet(pwd)

    def recover_p(self, priv, pwd):
        """
        通过私钥恢复
        """
        recover_api = BaseWalletAPI(init_account=True)
        return recover_api.recover_wallet(priv, pwd)

    def recover_m(self, mnemonic, pwd):
        """
        通过助记词恢复

        :param mnemonic: _description_
        :param pwd: _description_
        :return: _description_
        """
        wallet_api = BaseWalletAPI(init_account=True)
        return wallet_api.recover_m(mnemonic, pwd)



import os
from web3 import Web3
import time
import json
from pprint import pprint


def parse_path(path, num):
    for _ in range(1, num+1):
        path = path[:path.rfind("/")]
    return path


directory = os.path.dirname(__file__)
directory = parse_path(directory, 2)

with open(directory+"/conf/abi/abi.json", "r") as r:
    abi = json.loads(r.read())


class BaseEvent():
    def __init__(self) -> None:
        url = "https://ropsten.infura.io/v3/34ed41c4cf28406885f032930d670036"
        self.web3 = Web3(Web3.HTTPProvider(url))
        address = self.web3.toChecksumAddress(
            '0x43710f201183008EAf8170E729481289D52A695B')
        self.contract = self.web3.eth.contract(address=address, abi=abi)

    def main_function(self):
        latest = self.web3.eth.blockNumber
        transfer_events = self.contract.events.Transfer.createFilter(
            fromBlock=int(latest/5), toBlock=latest)
        data = transfer_events.get_all_entries()
        return data

    def loop_function(self):
        latest = self.web3.eth.blockNumber
        transfer_events = self.contract.events.Transfer.createFilter(
            fromBlock=latest-1, toBlock=latest)
        data = transfer_events.get_all_entries()
        return data

    def parsing_event(self, data):
        transfer_list = []
        for i in data:
            transfer_data = {}
            transfer_data['from'] = i['args']['from']
            transfer_data['to'] = i['args']['to']
            transfer_data['amount'] = i['args']['tokens']
            transfer_list.append(transfer_data)
        return transfer_list

    def get_transfer_events(self, realtime):
        data = self.main_function()
        pprint(self.parsing_event(data))
        print("↑↑↑↑↑↑↑↑↑↑↑↑↑↑历史数据↑↑↑↑↑↑↑↑↑↑↑↑↑↑")
        while 1:
            data = self.loop_function()
            if not data:
                print("监听中···")
            else:
                pprint(self.parsing_event(data))
            time.sleep(realtime)


if __name__ == "__main__":
    main = BaseEvent()
    main.get_transfer_events(10)



import yaml
import os

template = {
    "NETWORK": "",
    "KEYPATH": "",
    "KEYNAME": "",
    "GAS_URL": "",
    "TOKEN-LIST": [
        {"NAME": "", "ADDRESS": ""}
    ]
}


def parse_path(path, num):
    for _ in range(1, num+1):
        path = path[:path.rfind("/")]
    return path


directory = os.path.dirname(__file__)
directory = parse_path(directory, 2)


class ConfigUtil:
    def __init__(self):

        self.config_name = "config.yaml"
        self.config_path = directory + "/conf/"
        self.config_file = self.config_path+self.config_name

        if not os.path.exists(self.config_file):
            with open(self.config_file, 'w', encoding="utf-8") as w:
                yaml.safe_dump(template, w)

    def set(self, key, *value, **kwargs):
        data = self.get
        with open(self.config_file, 'w', encoding="utf-8") as w:
            if kwargs != {}:
                data[key].append(kwargs)
            else:
                data[key] = value
            yaml.safe_dump(data, w)
            return True

    @ property
    def get(self):
        with open(self.config_file, "r", encoding="utf-8") as r:
            return yaml.safe_load(r.read())


if __name__ == "__main__":
    config = ConfigUtil()
